#include <SDL2/SDL.h>

#include <array>

#pragma once

// inspiration for GUI:
// https://github.com/eabrash/game-of-life/blob/master/basic-draw/main.cpp

enum class cell_state : uint8_t { DEAD = 0, ALIVE = 1 };

// Game window for visualization of GoL, window will self-initialize.
// You only have to use the `draw_Cell_matrix` method when you want to 
// visualize actuall cell matrix.
//
// CellsWidth - width of the cell matrix - number of columns
// CellsHeight - height of the cell matrix - number of rows
// Scale - each cell is visualized as pixel square with ScalexScale size
//
template <std::size_t CellsWidth, std::size_t CellsHeight, std::size_t Scale>
class game_window {
	SDL_Window* window_ptr_ = nullptr;
	SDL_Surface* surface_ptr_ = nullptr;

       public:
	static constexpr std::size_t cells_width = CellsWidth;
	static constexpr std::size_t cells_height = CellsHeight;
	static constexpr std::size_t scale = Scale;
	static constexpr std::size_t window_width = cells_width * scale;
	static constexpr std::size_t window_height = cells_height * scale;

	static constexpr uint32_t alive_pixel = 0x00FFA400;
	static constexpr uint32_t dead_pixel = 0x00232528;

	using cell_row = std::array<cell_state, cells_width>;
	using cell_matrix = std::array<cell_row, cells_height>;

       public:
	game_window() {
		window_ptr_ =
		    SDL_CreateWindow("Game of life", SDL_WINDOWPOS_UNDEFINED,
				     SDL_WINDOWPOS_UNDEFINED, window_width,
				     window_height, SDL_WINDOW_SHOWN);
		if (window_ptr_ == nullptr) {
			std::cerr << "Failed to create SDL window";
			std::exit(2);
		}
		surface_ptr_ = SDL_GetWindowSurface(window_ptr_);

		SDL_memset(surface_ptr_->pixels, 0,
			   surface_ptr_->h * surface_ptr_->pitch);
	}

	game_window(const game_window&) = delete;
	game_window(game_window&&) = delete;
	game_window& operator=(const game_window&) = delete;
	game_window& operator=(game_window&&) = delete;

	void draw_cell_matrix(const cell_matrix& cells) {
		surface_ptr_ = SDL_GetWindowSurface(window_ptr_);
		for (std::size_t row_i = 0; row_i < window_height; ++row_i) {
			for (std::size_t col_i = 0; col_i < window_width;
			     ++col_i) {
				cell_state state =
				    cells[row_i / scale][col_i / scale];
				draw_pixel(row_i, col_i, state);
			}
		}
		SDL_UpdateWindowSurface(window_ptr_);
	}

	~game_window() {
		if (window_ptr_ == nullptr) {
			return;
		}
		SDL_DestroyWindow(window_ptr_);
	}

       private:
	void draw_pixel(std::size_t row_i, std::size_t col_i,
			cell_state state) {
		uint32_t* pixel_ptr =
		    reinterpret_cast<uint32_t*>(surface_ptr_->pixels) + col_i +
		    row_i * window_width;

		if (state == cell_state::ALIVE) {
			*pixel_ptr = alive_pixel;
		} else {
			*pixel_ptr = dead_pixel;
		}
	}
};
