#pragma once

#include <vector>
#include <memory>
#include <functional>
#include <unordered_map>

enum class NodeColor { WHITE, GRAY, BLACK };

struct Nothing {};

template < typename NodeData, typename EdgeData >
class Node {
public:
    struct Edge {
        Edge( Node *dest, const EdgeData& data ) :
            dest( dest ), data( data )
        {}

        Node *dest;
        EdgeData data;
    };

    Node( const NodeData& data ):
        data( data )
    {}

    const std::vector< Edge >& neighbors() const {
        return _neighbors;
    }

    void addNeighbor( Node* neighbor, const EdgeData& data ) {
        _neighbors.emplace_back( neighbor, data );
    }

    void addNeighbor( Node* neighbor ) {
        _neighbors.emplace_back( neighbor, EdgeData() );
    }

    NodeData data;
private:
    std::vector< Edge > _neighbors;
};

template < typename NodeData, typename EdgeData >
class Graph {
public:
    using NodeT = Node< NodeData, EdgeData >;

    NodeT *addNode( const NodeData& data = NodeData() ) {
        _nodes.emplace_back( std::make_unique< NodeT >( data ) );
        return _nodes.back().get();
    }

    const std::vector< std::unique_ptr< NodeT > >& nodes() const {
        return _nodes;
    }

    void walkDfs( std::function< void(NodeT&) > pre, std::function< void(NodeT&) > post ) {
        std::unordered_map< NodeT*, NodeColor > nodeColors;
        // Note that the default constructed color is white
        for ( const auto& node : _nodes ) {
            _runDfs( node.get(), nodeColors, pre, post );
        }
    }

private:
    void _runDfs( NodeT *node, std::unordered_map< NodeT*, NodeColor >& nodeColors,
        std::function< void(NodeT&) > pre, std::function< void(NodeT&) > post )
    {
        if ( nodeColors[ node ] != NodeColor::WHITE )
            return;
        pre( *node );
        nodeColors[ node ] = NodeColor::GRAY;
        for ( const auto& edge : node->neighbors() )
            _runDfs( edge.dest, nodeColors, pre, post );
        nodeColors[ node ] = NodeColor::BLACK;
        post( *node );
    }

    std::vector< std::unique_ptr< NodeT > > _nodes;
};